﻿using UnityEngine;
using System.Collections;
using System.Collections.Generic;
using Barsalt.ModularParameterEditor.EntryElements;

namespace Barsalt.ModularParameterEditor {
	/// <summary>
	/// The GroupElement manages and displays the EntryElements. 
	/// </summary>
	public sealed class GroupElement {
		#region CLASS_ATTRIBUTES
		//==========================================================================
		private static readonly int BORDER_HEIGHT		= 32;
		private static readonly int ELEMENT_HEIGHT		= 24;
		private static readonly int BUTTON_WIDTH		= 64;
		private static readonly int HEADER_HEIGHT		= 32;
		private static readonly int PADDING				= 4;
		private static readonly int SEARCHBAR_HEIGHT	= 32;
		//================================p==========================================
		#endregion
	
		#region OBJECT_ATTRIBUTES
		//==========================================================================
		private int size											= 0;
		private readonly List<EntryElement> ENTRY_LIST				= new List<EntryElement>();
		private readonly Dictionary<string, EntryElement> ENTRIES	= new Dictionary<string,EntryElement>();
		private readonly string key;
		private readonly ParameterEditor.Manager MANAGER;
		private string newEntryKey									= "";
		private string valueSearchKey								= "";
		private int parameterTypeIndex								= 0;
		private bool isEnabled										= true;
		//==========================================================================
		#endregion
	
		#region CONSTRUCTOR
		//==========================================================================
		/// <summary>
		/// 
		/// </summary>
		/// <param name="manager">The Managar is the interface for ParameterArchive modificatons. It is provided by the ParameterEditor</param>
		/// <param name="key">The key is the name of the Group that the GroupElement contains.</param>
		public GroupElement(ParameterEditor.Manager manager, string key) {
			this.MANAGER	= manager;
			this.key		= key;
		}
		//==========================================================================
		#endregion
	
		#region OBJECT_FUNCTIONS
		//==========================================================================
		/// <summary>
		/// The functions display the GroupElement and the containing EntryElements.
		/// </summary>The x position were the Element is to be displayed</param>
		/// <param name="y">The y position were the Element is to be displayed</param>
		/// <param name="width">The width of the Element</param>
		public void OnGui(int x, int y, int width){
			UpdateSize();

			#region HEADER_ELEMENT
			//==========================================================================
			GUI.Box(new Rect(x, y, width, Size), "");

			isEnabled = GUI.Toggle(new Rect(16, y + PADDING, ELEMENT_HEIGHT, ELEMENT_HEIGHT), isEnabled, "");

			GUI.Label(new Rect(x + 32, y + PADDING, 662, 32), key);

			if(GUI.Button(new Rect(x + width - PADDING - BUTTON_WIDTH, y + PADDING, BUTTON_WIDTH, ELEMENT_HEIGHT), "Remove")) {
				MANAGER.RemoveGroup(key);
			}
	
			y += HEADER_HEIGHT;
			//==========================================================================
			#endregion
	
			if(isEnabled){
				int entryWidth = width - 2 * PADDING;

				#region SEARCHBAR_ELEMENT
				//==========================================================================
				GUI.Box(new Rect(12, y, entryWidth, BORDER_HEIGHT), "");
				GUI.Box(new Rect(16, y + PADDING, BUTTON_WIDTH, ELEMENT_HEIGHT), "Search");

				valueSearchKey = GUI.TextField(new Rect(84, y + PADDING, width - 5 * PADDING - BUTTON_WIDTH, ELEMENT_HEIGHT), valueSearchKey);
	
				y += SEARCHBAR_HEIGHT + PADDING;
				//==========================================================================
				#endregion
	
				#region ENTRY_ELEMENTS
				//==========================================================================
				for(int index = 0; index < ENTRY_LIST.Count; index++) {
					if(valueSearchKey.Equals("")) {
						ENTRY_LIST[index].OnGui(12, y, entryWidth);
				
						if(ENTRY_LIST.Count > index){
							y += ENTRY_LIST[index].Size + PADDING;
						}
					} else if(ENTRY_LIST[index].Key.Contains(valueSearchKey)) {
						ENTRY_LIST[index].OnGui(12, y, entryWidth);
				
						if(ENTRY_LIST.Count > index) {
							y += ENTRY_LIST[index].Size + PADDING;
						}
					}
				}
				//==========================================================================
				#endregion
	
				#region ADD_ELEMENT
				//==========================================================================
				if(MANAGER.ContainsGroup(newEntryKey) || newEntryKey.Equals("") || MANAGER.ParameterTypeCount == 0)
					GUI.enabled = false;
	
				if(GUI.Button(new Rect(16, y, BUTTON_WIDTH, ELEMENT_HEIGHT), "Add")) {
					if(MANAGER.AddEntry(key, newEntryKey, MANAGER.GetParameterType(parameterTypeIndex).Default)) {
						Add(MANAGER.GetParameterType(parameterTypeIndex).NewEntryElement(MANAGER, Key, newEntryKey, MANAGER.GetParameterType(parameterTypeIndex).Default));
	
						newEntryKey = "";
					}
				}
	
				GUI.enabled = true;

				newEntryKey = GUI.TextField(new Rect(84, y, width - 2 * ELEMENT_HEIGHT - 10 * PADDING - BUTTON_WIDTH * 2, ELEMENT_HEIGHT), newEntryKey);
	
				if(MANAGER.ParameterTypeCount == 0)
					GUI.enabled = false;

				if(GUI.Button(new Rect(x + width - 2 * ELEMENT_HEIGHT - 6 * PADDING - BUTTON_WIDTH, y, ELEMENT_HEIGHT, ELEMENT_HEIGHT), "<")) {
					parameterTypeIndex--;
	
					if(parameterTypeIndex < 0) {
						parameterTypeIndex = MANAGER.ParameterTypeCount - 1;
					}
				}

				if(MANAGER.ParameterTypeCount == 0){
					GUI.Box(new Rect(x + width - ELEMENT_HEIGHT - 5 * PADDING - BUTTON_WIDTH - 2, y, 76, ELEMENT_HEIGHT), "none");
				} else {
					GUI.Box(new Rect(x + width - ELEMENT_HEIGHT - 5 * PADDING - BUTTON_WIDTH - 2, y, 76, ELEMENT_HEIGHT), MANAGER.GetParameterType(parameterTypeIndex).Name);
				}

				if(GUI.Button(new Rect(x + width - ELEMENT_HEIGHT - 2 * PADDING, y, ELEMENT_HEIGHT, ELEMENT_HEIGHT), ">")) {
					parameterTypeIndex++;
	
					if(parameterTypeIndex == MANAGER.ParameterTypeCount) {
						parameterTypeIndex = 0;
					}
				}
	
				GUI.enabled = true;
				//==========================================================================
				#endregion
			}
		}
	
		/// <summary>
		/// The function adds an EntryElement to the GroupElement.
		/// </summary>
		/// <param name="item">The EntryElement to be added</param>
		public void Add(EntryElement item){
			ENTRY_LIST.Add(item);
		
			ENTRIES.Add(item.Key, item);
		}
	
		/// <summary>
		/// The function replaces an old EntryElement with another one.
		/// </summary>
		/// <param name="key">The key of the EntryElement</param>
		/// <param name="item">The replacement EntryElement</param>
		public void Update(string key, EntryElement item){
			if(ENTRIES.ContainsKey(key)){
				int index = ENTRY_LIST.IndexOf(ENTRIES[key]);

				ENTRY_LIST.RemoveAt(index);

				ENTRY_LIST.Insert(index, item);

				ENTRIES[key] = item;
			}
		}

		/// <summary>
		/// The function removes an EntryElement from the GroupElement.
		/// </summary>
		/// <param name="key">The key of the EntryElement that should be removed</param>
		public void Remove(string key){
			ENTRY_LIST.Remove(ENTRIES[key]);
	
			ENTRIES.Remove(key);
		}

		private void UpdateSize(){
			size = 4 * PADDING + 2 * ELEMENT_HEIGHT + SEARCHBAR_HEIGHT;
	
			foreach(EntryElement element in ENTRY_LIST){
				if(element.Key.Contains(valueSearchKey)){
					size += element.Size + 4;
				}
			}
		}
		//==========================================================================
		#endregion
	
		#region GETTER_AND_SETTER
		//==========================================================================
		/// <summary>
		/// The Size of the GroupElement, based on its own elements and the size of the containing EntryElements
		/// </summary>
		public int Size{
			get{
				if(isEnabled){
					return size;
				} else {
					return 32;
				}
			}
		}
	
		/// <summary>
		/// The key of the group that is displayed in the GroupElement
		/// </summary>
		public string Key{
			get{
				return key;
			}
		}
		//==========================================================================
		#endregion
	}
}
