﻿using Barsalt.ModularParameterEditor.EntryElements;
using Barsalt.ModularParameterEditor.ParameterTypes;
using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using UnityEngine;
using Barsalt.WindowSystem;
using Barsalt.ParameterSystem;

namespace Barsalt.ModularParameterEditor {
	/// <summary>
	/// The ParameterEditor can load a ParameterArchive and display its contents. It can also be used to manipulate the entries of the ParameterArchive. If the ParameterEditor can manipulate a certain entry is based the the ParameterTypes that the ParameterEditor has.
	/// </summary>
	public sealed class ParameterEditor : BaseWindow {
		#region MANAGER_CLASS
		//==========================================================================
		/// <summary>
		/// The Manager is created by the ParameterEditor and provides and interface for the GroupElement and EntryElement to inform the Editor of changes to them.
		/// </summary>
		public sealed class Manager{
			#region OBJECT_ATTRIBUTES
			//==========================================================================
			private readonly ParameterEditor EDITOR;
			//==========================================================================
			#endregion

			#region CONSTRUCTOR
			//==========================================================================
			public Manager(ParameterEditor editor){
				this.EDITOR = editor;
			}
			//==========================================================================
			#endregion

			#region OBJECT_FUNCTIONS
			//==========================================================================
			/// <summary>
			/// The function informs the ParameterEditor that a new group has to be added to the ParameterArchive. 
			/// </summary>
			/// <param name="key">The name of the new group</param>
			/// <returns>Group was added successfully [true] or not [false]</returns>
			public bool AddGroup(string key){
				if(EDITOR.AddGroup(key)){
					return true;
				} else {
					return false;
				}
			}

			/// <summary>
			/// The function checks if a group with the passed key exists in the ParameterArchive.
			/// </summary>
			/// <param name="key">The key of the group</param>
			/// <returns>Group exists in the ParameterArchive [true] or not [false]</returns>
			public bool ContainsGroup(string key){
				return(EDITOR.archive.ContainsGroup(key));
			}

			/// <summary>
			/// The function informs the ParameterEditor that a group has to be removed from the ParameterArchive.
			/// </summary>
			/// <param name="key">The key of the group</param>
			/// <returns>Group was removed succesfully [true] or not [false]</returns>
			public bool RemoveGroup(string key){
				if(EDITOR.RemoveGroup(key)){
					return true;
				} else {
					return false;
				}
			}

			/// <summary>
			/// The function informs the ParameterEditor that a entry has to be added to the ParameterArchive
			/// </summary>
			/// <param name="group">The key of the group</param>
			/// <param name="key">The key of the entry</param>
			/// <param name="value">The value of the entry</param>
			/// <returns>Entry was added succesfully [true] or not [false]</returns>
			public bool AddEntry(string group, string key, object value){
				if(EDITOR.AddEntry(group, key, value)){
					return true;
				} else {
					return false;
				}
			}

			/// <summary>
			/// The function informs the ParameterEditor that a entry has to be updated in the ParameterArchive.
			/// </summary>
			/// <param name="group">The key of the group</param>
			/// <param name="key">The key of the entry</param>
			/// <param name="value">The value of the entry</param>
			/// <returns>Entry was updated succesfully [true] or not [false]</returns>
			public bool UpdateEntry(string group, string key, object value) {
				if(EDITOR.UpdateEntry(group, key, value)){
					return true;
				} else {
					return false;
				}
			}

			/// <summary>
			/// The function informs the ParameterEditor that a entry has to be removed from the ParameterArchive.
			/// </summary>
			/// <param name="group">The key of the group</param>
			/// <param name="key">The key of the entry</param>
			/// <returns>Entry was removed succesfully [true] or not [false]</returns>
			public bool RemoveEntry(string group, string key) {
				if(EDITOR.RemoveEntry(group, key)){
					return true;
				} else {
					return false;
				}
			}
			//==========================================================================
			#endregion

			#region GETTER_AND_SETTER
			//==========================================================================
			/// <summary>
			/// The number of ParameterTypes that the ParameerEditor has.
			/// </summary>
			public int ParameterTypeCount{
				get{
					return EDITOR.PARAMETER_TYPES.Count;
				}
			}

			/// <summary>
			/// Retrieve a certain ParameterType based on its position in the list.
			/// </summary>
			/// <param name="index">Position of the ParameterType in the list</param>
			/// <returns>The ParameterType</returns>
			public BaseParameterType GetParameterType(int index){
				if(EDITOR.PARAMETER_TYPES.Count > index){
					return EDITOR.PARAMETER_TYPE_LIST[index];
				} else {
					return null;
				}
			}
			//==========================================================================
			#endregion
		}
		//==========================================================================
		#endregion

		#region EDITOR_DELEGATE_DECLARATION
		//==========================================================================
		/// <summary>
		/// Delegate for the function that loads the ParameterArchive into the ParameterEditor. 
		/// </summary>
		/// <returns>The ParameterArchive that the ParamterEditor loads</returns>
		public delegate ParameterArchive ArchiveLoadCallback();

		/// <summary>
		/// Delegate for the function that saves the ParameterArchive that is loaded by the ParameterEditor.
		/// </summary>
		/// <param name="archive">The ParameterArchive that is loaded into the ParameterEditor</param>
		/// <returns></returns>
		public delegate bool ArchiveSaveCallback(ParameterArchive archive);
		//==========================================================================
		#endregion

		#region CLASS_ATTRIBUTES
		//==========================================================================
		public static readonly int BORDER_HEIGHT	= 32;
		public static readonly int ELEMENT_HEIGHT	= 24;
		public static readonly int PADDING			= 4;
		public static readonly int BUTTON_WIDTH		= 64;
		private static readonly int HEADER_HEIGHT	= 40;
		private static readonly int TITLEBAR_HEIGHT	= 20;
		//==========================================================================
		#endregion

		#region OBJECT_MANAGEMENT_ATTRIBUTES
		//==========================================================================
		private ParameterArchive archive			= null; 
		private ArchiveLoadCallback archiveLoader;
		private ArchiveSaveCallback archiveSaver;
		private string newGroupKey					= "";
		private string groupSearchkey				= "";
		private readonly Manager MANAGER;
		//==========================================================================
		#endregion

		#region OBJECT_GUI_ATTRIBUTES
		//==========================================================================
		private Rect menuBackground												= new Rect(4, 20, 760, 40);
		private Rect loadButton													= new Rect(138, 24, 162, 32);
		private Rect saveButton													= new Rect(303, 24, 162, 32);
		private Rect distributeButton											= new Rect(469, 24, 162, 32);
		private Rect groupSearchBackground										= new Rect(4, 64, 760, 32);
		private Rect groupSearchLabel											= new Rect(8, 68, 64, 24);
		private Rect groupSearchField											= new Rect(76, 68, 684, 24);

		private Rect parameterArea												= new Rect(4, 100, 760, 920);
		private Rect parameterAreaSize											= new Rect(4, 64, 744, 2000);
		private Rect noArchiveLoadedLabel										= new Rect(320, 530, 128, 32);
		private int scrollAreaHeight = 0;
		private Vector2 parameterAreaPosition									= Vector2.zero;

		private readonly List<GroupElement> GROUP_ELEMENT_LIST					= new List<GroupElement>();
		private readonly Dictionary<string, GroupElement> GROUP_ELEMENTS		= new Dictionary<string,GroupElement>();

		private readonly Dictionary<Type, BaseParameterType> PARAMETER_TYPES	= new Dictionary<Type, BaseParameterType>();
		private readonly List<BaseParameterType> PARAMETER_TYPE_LIST			= new List<BaseParameterType>();
		//==========================================================================
		#endregion

		#region CONSTRUCTOR
		//==========================================================================
		/// <summary>
		/// Creates the ParameterEditor and prepares the necessary elemtns like the Manager.
		/// </summary>
		/// <param name="x">The x position of the window</param>
		/// <param name="y">The y position of the window</param>
		/// <param name="title">The Title of the window</param>
		public ParameterEditor(int x, int y, string title) : base(new Rect(x, y, 768, 1024), title){
			MANAGER		= new Manager(this);

			MinimalSize = new Vector2(512, 512);
			MaximalSize = new Vector2(2048, 2048);
		}
		//==========================================================================
		#endregion

		#region OBJECT_FUNCTIONS
		//==========================================================================
		protected override void OnGui(int id) {
			OnGuiMenu();

			OnGuiContent();
		}

		private void OnGuiMenu(){
			GUI.Box(menuBackground, "");
			
			if(archiveLoader == null)
				GUI.enabled = false;

			if(GUI.Button(loadButton, "Load Archive")){
				Archive = archiveLoader();
			}

			GUI.enabled = true;

			if(archiveSaver == null || archive == null)
				GUI.enabled = false;

			if(GUI.Button(saveButton, "Save Archive")) {
				Save();
			}

			GUI.enabled = true;

			if(archive == null)
				GUI.enabled = false;

			if(GUI.Button(distributeButton, "Redistribute")){
				Distribute();
			}

			GUI.enabled = true;

			GUI.Box(groupSearchBackground, "");

			if(archive == null)
				GUI.enabled = false;

			GUI.Box(groupSearchLabel, "Search");
			groupSearchkey = GUI.TextField(groupSearchField, groupSearchkey);

			GUI.enabled = true;
		}

		private void OnGuiContent() {
			UpdateSize();

			parameterAreaSize = new Rect(parameterAreaSize.x, parameterAreaSize.y, parameterAreaSize.width, Math.Max(parameterArea.height, scrollAreaHeight));

			GUI.Box(parameterArea, "");
			
			int position	= 68;
			int width		= (int) parameterAreaSize.width - 2 * PADDING;

			if(archive != null) {
				parameterAreaPosition = GUI.BeginScrollView(parameterArea, parameterAreaPosition, parameterAreaSize, false, true);

				for(int index = 0; index < GROUP_ELEMENT_LIST.Count; index++){
					if(groupSearchkey.Equals("")){
						GROUP_ELEMENT_LIST[index].OnGui(8, position, width);

						if(index < GROUP_ELEMENT_LIST.Count){
							position += GROUP_ELEMENT_LIST[index].Size + 4;
						}
					} else if( GROUP_ELEMENT_LIST[index].Key.Contains(groupSearchkey)){
						GROUP_ELEMENT_LIST[index].OnGui(8, position, width);

						if(index < GROUP_ELEMENT_LIST.Count){
							position += GROUP_ELEMENT_LIST[index].Size + 4;
						}
					}
				}

				GUI.Box(new Rect(8, position, width, HEADER_HEIGHT), "");

				position += 4;

				GUI.Box(new Rect(12, position, width - 2 * PADDING, BORDER_HEIGHT), "");

				position += 4;

				if(newGroupKey == null || newGroupKey.Equals("") || archive.ContainsGroup(newGroupKey))
					GUI.enabled = false;

				if(GUI.Button(new Rect(16, position, BUTTON_WIDTH, ELEMENT_HEIGHT), "Add")){
					AddGroup(newGroupKey);

					newGroupKey = "";
				}

				GUI.enabled = true;

				newGroupKey = GUI.TextField(new Rect(84, position, width - BUTTON_WIDTH - 5 * PADDING, ELEMENT_HEIGHT), newGroupKey);

				GUI.EndScrollView();
			} else {
				GUI.Label(noArchiveLoadedLabel, "No Archive Loaded");
			}
		}

		private void Save(){
			archiveSaver(archive);
		}

		private void Distribute(){
			if(archive != null){
				archive.CallListeners();
			}
		}

		/// <summary>
		/// The function adds a new ParameterType to the editors selection and replaces all EntryElements of parameters with corresponding Types with the new EntryElement.
		/// </summary>
		/// <param name="type">The new ParameterType</param>
		public void Add(BaseParameterType type) {
			if(!PARAMETER_TYPES.ContainsKey(type.type)) {
				PARAMETER_TYPES.Add(type.type, type);

				PARAMETER_TYPE_LIST.Add(type);

				if(archive != null){
					foreach(string group in archive.Keys()) {
						foreach(string key in archive.Keys(group)) {
							object value = archive.GetValue<object>(group, key);

							if(value.GetType() == type.type) {
								GROUP_ELEMENTS[group].Update(key, PARAMETER_TYPES[value.GetType()].NewEntryElement(MANAGER, group, key, value));
							}
						}
					}
				}
			}
		}

		/// <summary>
		/// The function removes a ParameterType from the editor and replaces all EntryElements of parameters with corresponding Types with NullEntryElements.
		/// </summary>
		/// <param name="type">The ParameterType that is to be removed</param>
		public void Remove(BaseParameterType type) {
			if(PARAMETER_TYPES.ContainsKey(type.type)) {
				PARAMETER_TYPES.Remove(type.type);

				PARAMETER_TYPE_LIST.Remove(type);

				if(archive != null){
					foreach(string group in archive.Keys()) {
						foreach(string key in archive.Keys(group)) {
							object value = archive.GetValue<object>(group, key);

							if(value.GetType() == type.type){
								GROUP_ELEMENTS[group].Update(key, new NullEntryElement(MANAGER, group, key, value));
							}
						}
					}
				}
			}
		}

		private bool AddGroup(string key){
			if(archive.AddGroup(key)){
			
				GroupElement group = new GroupElement(MANAGER, key);

				GROUP_ELEMENT_LIST.Add(group);

				GROUP_ELEMENTS.Add(key, group);

				return true;
			} else {
				return false;
			}
		}

		private bool ContainsGroup(string key) {
			return archive.ContainsGroup(key);
		}

		private bool RemoveGroup(string key) {
			if(archive.RemoveGroup(key)){
				GROUP_ELEMENT_LIST.Remove(GROUP_ELEMENTS[key]);

				GROUP_ELEMENTS.Remove(key);
	
				return true;
			} else {
				return false;
			}
		}

		private bool AddEntry(string group, string key, object value) {
			if(archive.AddValue(group, key, value)){
				return true;
			} else {
				return false;
			}
		}

		private bool UpdateEntry(string group, string key, object value){
			return archive.UpdateValue(group, key, value);
		}

		private bool RemoveEntry(string group, string key) {
			if(archive.RemoveValue(group, key)){
				GROUP_ELEMENTS[group].Remove(key);

				return true;
			} else {
				return false;
			}
		}

		private void UpdateSize() {
			scrollAreaHeight = PADDING + 40 + PADDING;

			foreach(GroupElement element in GROUP_ELEMENT_LIST) {
				scrollAreaHeight += element.Size + PADDING;
			}
		}

		private void SetupEditor() {
			if(archive != null){
				GROUP_ELEMENT_LIST.Clear();
				GROUP_ELEMENTS.Clear();

				foreach(string group in archive.Keys()) {
					GroupElement element = new GroupElement(MANAGER, group);

					foreach(string key in archive.Keys(group)) {
						object value = archive.GetValue<object>(group, key);

						
						if(PARAMETER_TYPES.ContainsKey(value.GetType())){
							element.Add(PARAMETER_TYPES[value.GetType()].NewEntryElement(MANAGER, group, key, value));
						} else {
							element.Add(new NullEntryElement(MANAGER, group, key, value));
						}
					}

					GROUP_ELEMENT_LIST.Add(element);

					GROUP_ELEMENTS.Add(group, element);

				}
			}
		}

		protected override void OnDimensionChange(){
			int buttonWidth			= (int)(Dimension.width - 6 * PADDING) / 3;
			int currentHeight		= TITLEBAR_HEIGHT;

			menuBackground			= new Rect(PADDING, currentHeight, Dimension.width - 2 * PADDING, HEADER_HEIGHT);

			currentHeight += PADDING;
			loadButton				= new Rect(2 * PADDING, currentHeight, buttonWidth, BORDER_HEIGHT);
			saveButton				= new Rect(3 * PADDING + buttonWidth, currentHeight, buttonWidth, BORDER_HEIGHT);
			distributeButton		= new Rect(4 * PADDING + buttonWidth * 2, currentHeight, buttonWidth, BORDER_HEIGHT);

			currentHeight += HEADER_HEIGHT;
			groupSearchBackground	= new Rect(PADDING, currentHeight, Dimension.width - 2* PADDING, BORDER_HEIGHT);

			currentHeight += PADDING;
			groupSearchLabel		= new Rect(2 * PADDING, currentHeight, BUTTON_WIDTH, ELEMENT_HEIGHT);
			groupSearchField		= new Rect(3 * PADDING + BUTTON_WIDTH, currentHeight, Dimension.width - 5 * PADDING - BUTTON_WIDTH, ELEMENT_HEIGHT);

			currentHeight += PADDING + ELEMENT_HEIGHT + PADDING;
			parameterArea			= new Rect(PADDING, currentHeight, Dimension.width - 2 * PADDING, Dimension.height - currentHeight - PADDING);
			parameterAreaSize		= new Rect(PADDING, 64, Dimension.width - ELEMENT_HEIGHT, 2000);
			noArchiveLoadedLabel	= new Rect(Dimension.width / 2 - BUTTON_WIDTH, parameterArea.height / 2 - BORDER_HEIGHT / 2 + currentHeight, BUTTON_WIDTH * 2, BORDER_HEIGHT);
		}
		//==========================================================================
		#endregion

		#region GETTER_AND_SETTER
		//==========================================================================
		/// <summary>
		/// The callback function to load a ParameterArchive in the ParameterEditor
		/// </summary>
		public ArchiveLoadCallback ArchiveLoader{
			set{
				archiveLoader = value;
			}
		}

		/// <summary>
		/// The callback function to save a ParameterArchive
		/// </summary>
		public ArchiveSaveCallback ArchiveSaver {
			set {
				archiveSaver = value;
			}
		}

		/// <summary>
		/// The ParameterArchive that is currently assigned to the ParameterEditor 
		/// </summary>
		public ParameterArchive Archive{
			get{
				return archive;
			}
			set{
				if(value != null){
					archive = value;

					SetupEditor();
				}
			}
		}
		//==========================================================================
		#endregion
	}
}
