﻿using UnityEngine;
using System.Collections;
using System;
using System.Collections.Generic;
using Barsalt.ParameterSystem.Container;
using Barsalt.ParameterSystem.ContainerProvider;
using Barsalt.ParameterSystem.Exceptions;
using Assets.ParameterSystem;
using System.Xml.Serialization;
using System.Xml;

namespace Barsalt.ParameterSystem {
	/// <summary>
	/// <para>The ParameterArchive provides a group and key-value pair based storrage for any kind of object.</para>
	/// <para>Each ParameterArchive allows the registration of ParameterListeners that will be informed of changes in the ParameterArchive.</para>
	/// 
	/// <para> - To inform the ParameterListener of any changes it is necessary to call the CallListeners() function of the ParameterArchive.</para>
	/// <para> - The ParameterArchive supports Binary- and XML-Serialization.</para>
	/// <para> - The ParameterArchive uses the OnDistribution() function of the group to retrieve the keys of entries that need to be distributed to the ParameterListener.</para>
	/// </summary>
	[Serializable()]
	public sealed class ParameterArchive : IXmlSerializable{
		#region OBJECT_ATTRIBUTES
		//==========================================================================
		[NonSerialized()]
		private readonly Dictionary<string, HashSet<ParameterListener>> LISTENERS = new Dictionary<string,HashSet<ParameterListener>>();
		private readonly Dictionary<string, IParameterMap> GROUPS	= new Dictionary<string, IParameterMap>();
		private IParameterMapProvider mapProvider					= new ParameterGroupProvider();
		private string name											= "";
		//==========================================================================
		#endregion

		#region CONSTRUCTORS
		//==========================================================================
		public ParameterArchive() : this(new ParameterGroupProvider()) {}

		/// <summary>
		/// Constructor that takes a IParameterMapProvider to change which implementation of the IParameterMap is used. 
		/// </summary>
		/// <param name="mapProvider">The provider for the groups</param>
		public ParameterArchive(IParameterMapProvider mapProvider) {
			this.mapProvider	= mapProvider;
		}
		//==========================================================================
		#endregion
	
		#region OBJECT_FUNCTIONS
		//==========================================================================
		/// <summary>
		/// Adds a group with the key to the ParameterArchive. The group is created by the IParameterMapProvider.
		/// </summary>
		/// <param name="key">The key of the group</param>
		/// <returns>Returns true if the new group was added</returns>
		public bool AddGroup(string key){
			if(!GROUPS.ContainsKey(key)){
				GROUPS.Add(key, mapProvider.NewMap());
				LISTENERS.Add(key, new HashSet<ParameterListener>());

				return true;
			}

			throw new ParameterKeyException(key);
		}
	
		/// <summary>
		/// Removes the group with the key from the ParameterArchive.
		/// </summary>
		/// <param name="key">The key of the group</param>
		/// <returns>Returns true if the group was removed</returns>
		public bool RemoveGroup(string key){
			if(GROUPS.ContainsKey(key)){
				return (GROUPS.Remove(key) && LISTENERS.Remove(key));
			}

			throw new ParameterKeyException(key);
		}

		/// <summary>
		/// Adds an entry with the key and value into the group of the ParameterArchive. 
		/// </summary>
		/// <param name="group">The key of the group</param>
		/// <param name="key">The key of the entry</param>
		/// <param name="value">The value of the entry</param>
		/// <returns>Returns true if the entry was added</returns>
		public bool AddValue(string group, string key, object value){
			if(GROUPS.ContainsKey(group)){
				return GROUPS[group].Add(key, value);
			}

			throw new ParameterKeyException(group);
		}
	
		/// <summary>
		/// Updates the value of the entry identified by the group and entry key.
		/// </summary>
		/// <param name="group">The key of the group</param>
		/// <param name="key">The key of the entry</param>
		/// <param name="value">The value of the entry</param>
		/// <returns>Returns true if the entry was updated</returns>
		public bool UpdateValue(string group, string key, object value){
			if(GROUPS.ContainsKey(group)) {
				return GROUPS[group].Update(key, value);
			}

			throw new ParameterKeyException(group);
		}
		
		/// <summary>
		/// Removes the entry identified by the keys. 
		/// </summary>
		/// <param name="group">The key of the group</param>
		/// <param name="key">The key of the entry</param>
		/// <returns>Returns true if the entry was removed</returns>
		public bool RemoveValue(string group, string key){
			if(GROUPS.ContainsKey(group)) {
				return GROUPS[group].Remove(key);
			}

			throw new ParameterKeyException(group);
		}
	
		
		/// <summary>
		/// Checks if the ParameterArchive contains a group with the key.
		/// </summary>
		/// <param name="group">The key of the group</param>
		/// <returns>Returns true if the group exists</returns>
		public bool ContainsGroup(string group){
			return GROUPS.ContainsKey(group);
		}

		/// <summary>
		/// Calls all listeners to inform them that the ParameterArchive has been changed.
		/// </summary>
		public void CallListeners(){
			foreach(string group in GROUPS.Keys){
				List<string> keys = GROUPS[group].OnDistribution();

				if(keys.Count > 0){
					ParameterEvent events = new ParameterEvent(this, group, keys);

					foreach(ParameterListener listener in LISTENERS[group]) {
						listener.OnParameterChange(events);
					}
				}
			}
		}

		/// <summary>
		/// Adds the listener and makes an immediat callback if necessary.
		/// </summary>
		/// <param name="listener">The listener</param>
		/// <param name="callback">True for callback</param>
		public void AddParameterListener(string group, ParameterListener listener){
		//Debug.Log(); 
			if(LISTENERS.ContainsKey(group)){
				LISTENERS[group].Add(listener);
			} else {
				throw new ParameterKeyException(group);
			}
		}

		/// <summary>
		/// Removes the listener.
		/// </summary>
		/// <param name="listener">The listener</param>
		public void RemoveParameterListener(string group, ParameterListener listener) {
			LISTENERS[group].Remove(listener);
		}

		/// <summary>
		/// Deserializes the ParameterArchive from an XML stream.
		/// </summary>
		/// <param name="reader">The stream for the deserialization</param>
		public void ReadXml(XmlReader reader) {
			#region SETUP
			//==========================================================================
			XmlSerializer deserializer = null;
			List<string> keys = new List<string>();
			//==========================================================================
			#endregion

			reader.MoveToContent();

			name = reader.GetAttribute("name");

			int groupCount = int.Parse(reader.GetAttribute("count"));

			#region READ_MAP_PROVIDER
			//==========================================================================
			reader.ReadToFollowing("MapProvider");

			// prepare XmlSerializer with type based on MapProvider data
			deserializer = new XmlSerializer(Type.GetType(reader.GetAttribute("type") + ", " + reader.GetAttribute("assembly")));

			reader.Read();

			mapProvider = (IParameterMapProvider)deserializer.Deserialize(reader.ReadSubtree());
			//==========================================================================
			#endregion

			#region READ_KEYS
			//==========================================================================
			reader.ReadToFollowing("Key");

			for(int index = 0; index < groupCount; index++) {
				keys.Add(reader.GetAttribute("key"));

				reader.Read();
			}
			//==========================================================================
			#endregion

			#region READ_GROUPS
			//==========================================================================
			reader.ReadToFollowing("Groups");

			// step into first group element
			reader.Read();

			// retrieve name of node for call of ReadToFollowing
			string nodeName = reader.Name;

			for(int index = 0; index < groupCount; index++) {
				deserializer = new XmlSerializer(mapProvider.MapType);

				IParameterMap map = (IParameterMap)deserializer.Deserialize(reader.ReadSubtree());

				// add the map with the associated key into the dictionary
				GROUPS.Add(keys[index], map);
				LISTENERS.Add(keys[index], new HashSet<ParameterListener>());

				reader.ReadToFollowing(nodeName);
			}
			//==========================================================================
			#endregion
		}

		/// <summary>
		/// Serializes the ParameterArchive into a stream.
		/// </summary>
		/// <param name="writer">The stream for the serialization</param>
		public void WriteXml(XmlWriter writer) {
			#region SETUP
			//==========================================================================
			XmlSerializer serializer = null;
			//==========================================================================
			#endregion

			writer.WriteAttributeString("name", name);

			writer.WriteAttributeString("count", GROUPS.Count.ToString());

			#region WRITE_MAP_PROVIDER
			//==========================================================================
			writer.WriteStartElement("MapProvider");

			writer.WriteAttributeString("type", mapProvider.GetType().ToString());

			writer.WriteAttributeString("assembly", mapProvider.GetType().Assembly.ToString());

			serializer = new XmlSerializer(mapProvider.GetType());

			serializer.Serialize(writer, mapProvider);

			writer.WriteEndElement();
			//==========================================================================
			#endregion

			#region WRITE_KEYS
			//==========================================================================
			writer.WriteStartElement("Keys");

			string[] keys = new string[GROUPS.Count];

			GROUPS.Keys.CopyTo(keys, 0);

			foreach(string key in keys) {
				writer.WriteStartElement("Key");

				writer.WriteAttributeString("key", key);

				writer.WriteEndElement();
			}

			writer.WriteEndElement();
			//==========================================================================
			#endregion

			#region WRITE-GROUPS
			//==========================================================================
			writer.WriteStartElement("Groups");

			foreach(string key in keys) {

				serializer = new XmlSerializer(GROUPS[key].GetType());

				serializer.Serialize(writer, GROUPS[key]);
			}

			writer.WriteEndElement();
			//==========================================================================
			#endregion
		}
		//==========================================================================
		#endregion
	
		#region GETTER_AND_SETTER
		//==========================================================================
		public IParameterMapProvider ParameterMapProvider{
			get{
				return mapProvider;
			}
		}

		public string Name{
			get{
				return name;
			}
			set{
				name = value;
			}
		}

		/// <summary>
		/// 
		/// </summary>
		/// <param name="group">The key of the group</param>
		/// <returns>Returns a list of all keys in the group</returns>
		public List<string> Keys(string group) {
			if(GROUPS.ContainsKey(group)) {
				return GROUPS[group].Keys;
			}

			throw new ParameterKeyException(group);
		}

		/// <summary>
		/// 
		/// </summary>
		/// <returns>Returns a list of all group keys</returns>
		public List<string> Keys() {
			List<string> keys = new List<string>();

			foreach(string key in GROUPS.Keys) {
				keys.Add(key);
			}

			return keys;
		}

		/// <summary>
		///
		/// </summary>
		/// <param name="group">The key of the group</param>
		/// <param name="key">The key of the entry</param>
		/// <returns>Returns the Type of the value specified by the keys</returns>
		public Type IsType(string group, string key) {
			if(GROUPS.ContainsKey(group)) {
				return GROUPS[group].GetType(key);
			}

			throw new ParameterKeyException(group);
		}

		/// <summary>
		/// Returns the value as T.
		/// </summary>
		/// <typeparam name="T"></typeparam>
		/// <param name="group">The key of the group</param>
		/// <param name="key">The key of the entry</param>
		/// <returns>Returns the the value specified by the keys</returns>
		public T GetValue<T>(string group, string key) where T : new() {
			if(GROUPS.ContainsKey(group)) {
				return GROUPS[group].GetValue<T>(key);
			}

			throw new ParameterKeyException(group);
		}

		/// <summary>
		///
		/// </summary>
		/// <returns>Returns null because I was to lazy to read into XML-Schema and implement one</returns>
		public System.Xml.Schema.XmlSchema GetSchema() {
			return null;
		}
		//==========================================================================
		#endregion
	}
}