﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;

using UnityEngine;

namespace Barsalt.WindowSystem.Components {
	/// <summary>
	/// The IdSystem manages a pool of IDs. Each ID in the pool is unique. The system manages the size of the pool, enlarging it if necessary. The system provides functions to get IDs, return IDs and claim a certain ID.
	/// </summary>
	public class IdSystem {
		#region ID_CLAIM_DELEGATE
		//==========================================================================
		public delegate void ID_ClaimHandler(int id);
		//==========================================================================
		#endregion
	
		#region OBJECT_ATTRIBUTES
		//==========================================================================
		private readonly HashSet<int> avaiable = new HashSet<int>();
		private readonly HashSet<int> unavaiable = new HashSet<int>();
		private int poolSize;
		private ID_ClaimHandler changeHandler;
		//==========================================================================
		#endregion
	
		#region CONSTRUCTOR
		//==========================================================================
		/// <summary>
		/// 
		/// </summary>
		/// <param name="poolSize">The start size of the pool of IDs</param>
		public IdSystem(int poolSize) {
			this.poolSize = poolSize;

			for(int counter = 0; counter < poolSize; counter++) {
				avaiable.Add(GenerateID());
			}
		}
		//==========================================================================
		#endregion
	
		#region OBJECT_FUNCTIONS
		//==========================================================================
		/// <summary>
		/// Retrieve an ID from the pool.
		/// </summary>
		/// <returns></returns>
		public int GetID() {
			if(avaiable.Count > 0) {

				int id = avaiable.ElementAt<int>(0);

				avaiable.Remove(id);

				unavaiable.Add(id);

				return id;
			} else {
				int id = GenerateID();

				unavaiable.Add(id);

				return id;
			}
		}

		/// <summary>
		/// Return a ID to the pool.
		/// </summary>
		/// <param name="id"></param>
		public void ReturnID(int id) {
			if(unavaiable.Contains(id)) {
				unavaiable.Remove(id);

				if(!(avaiable.Count == poolSize))
					avaiable.Add(id);
			}
		}

		private int GenerateID() {
			int id = -1;
			int counter = 0;

			do {
				if(!avaiable.Contains(counter) && !unavaiable.Contains(counter))
					id = counter;
				else
					counter++;
			} while(id == -1);

			return id;
		}

		/// <summary>
		/// Claim a certain ID from the Manager.
		/// </summary>
		/// <param name="id"></param>
		public void ClaimID(int id) {
			if(avaiable.Contains(id)) {
				avaiable.Remove(id);

				unavaiable.Add(id);
			} else if(!avaiable.Contains(id) && !unavaiable.Contains(id)) {
				unavaiable.Add(id);
			} else if(unavaiable.Contains(id)) {
				changeHandler(id);
			}
		}

		/// <summary>
		/// ClaimListener are called when someone claims a certain ID so that they can check if their ID was claimed and act accordingly.
		/// </summary>
		/// <param name="handler"></param>
		public void AddClaimID_Handler(ID_ClaimHandler handler) {
			changeHandler += handler;
		}

		public void RemoveClaimID_Handler(ID_ClaimHandler handler) {
			changeHandler -= handler;
		}
		//==========================================================================
		#endregion
	
		#region GETTER_AND_SETTER
		//==========================================================================
		public int PoolSize {
			get {
				return poolSize;
			}
			set {
				poolSize = value;
			}
		}
		//==========================================================================
		#endregion
	}
}